if bytes is str:  # pragma: no cover
    def isStringType(t):
        return isinstance(t, basestring)

    def isPath(f):
        return isinstance(f, basestring)

else:  # pragma: no cover
    def isStringType(t):
        return isinstance(t, str)

    def isPath(f):
        return isinstance(f, (bytes, str))



from __future__ import division, unicode_literals, absolute_import

from itertools import chain

from . import Image, ImageFilter, ImageMath
from .utils import isPath

try:
    import numpy
except ImportError:  # pragma: no cover
    numpy = None


def load_cube_file(lines, target_mode=None, cls=ImageFilter.Color3DLUT):
    """Loads 3D lookup table from .cube file format.

    :param lines: Filename or iterable list of strings with file content.
    :param target_mode: Image mode which should be after color transformation.
                        The default is None, which means mode doesn't change.
    :param cls: A class which handles the parsed file.
                Default is ``ImageFilter.Color3DLUT``.
    """
    name, size = None, None
    channels = 3
    file = None

    if isPath(lines):
        file = lines = open(lines, 'rt')

    try:
        iterator = iter(lines)

        for i, line in enumerate(iterator, 1):
            line = line.strip()
            if line.startswith('TITLE "'):
                name = line.split('"')[1]
                continue
            if line.startswith('LUT_3D_SIZE '):
                size = [int(x) for x in line.split()[1:]]
                if len(size) == 1:
                    size = size[0]
                continue
            if line.startswith('CHANNELS '):
                channels = int(line.split()[1])
            if line.startswith('LUT_1D_SIZE '):
                raise ValueError("1D LUT cube files aren't supported")

            try:
                float(line.partition(' ')[0])
            except ValueError:
                pass
            else:
                # Data starts
                break

        if size is None:
            raise ValueError('No size found in the file')

        table = []
        for i, line in enumerate(chain([line], iterator), i):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            try:
                pixel = [float(x) for x in line.split()]
            except ValueError:
                raise ValueError("Not a number on line {}".format(i))
            if len(pixel) != channels:
                raise ValueError(
                    "Wrong number of colors on line {}".format(i))
            table.extend(pixel)
    finally:
        if file is not None:
            file.close()

    instance = cls(size, table, channels=channels,
                   target_mode=target_mode, _copy_table=False)
    if name is not None:
        instance.name = name
    return instance


def load_hald_image(image, target_mode=None, cls=ImageFilter.Color3DLUT):
    """Loads 3D lookup table from Hald image (normally .png or .tiff files).

    :param image: Pillow RGB image or path to the file.
    :param target_mode: Image mode which should be after color transformation.
                        The default is None, which means mode doesn't change.
    :param cls: A class which handles the parsed file.
                Default is ``ImageFilter.Color3DLUT``.
    """
    if not isinstance(image, Image.Image):
        image = Image.open(image)

    if image.size[0] != image.size[1]:
        raise ValueError("Hald image should be a square")

    channels = len(image.getbands())

    for i in range(2, 9):
        if image.size[0] == i**3:
            size = i**2
            break
    else:
        raise ValueError("Can't detect hald size")

    if numpy:
        table = numpy.array(image).reshape(size**3 * channels)
        table = table.astype(numpy.float32) / 255.0
    else:
        table = []
        for color in zip(*[
            ImageMath.eval("a/255.0", a=im.convert('F')).im
            for im in image.split()
        ]):
            table.extend(color)

    return cls(size, table, target_mode=target_mode, _copy_table=False)



#!/usr/bin/env python

from setuptools import setup

setup(
    name='pillow_lut',
    version='1.0',
    description='Lookup tables loading, manipulation and generation '
                'for Pillow library',
    author='Alex Karpinsky',
    author_email='homm86@gmail.com',
    url='https://github.com/homm/pillow-lut-tools',
    classifiers=[
        "Development Status :: 6 - Mature",
        "Topic :: Multimedia :: Graphics",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    packages=['pillow_lut'],
    setup_requires=['nose>=1.3']
)



from __future__ import division, unicode_literals, absolute_import

from math import sin

from . import ImageFilter

try:
    import numpy
except ImportError:  # pragma: no cover
    numpy = None


def _srgb_to_linear(s):
    if s < 0.0404482362771082:
        return s / 12.92
    return pow((s + 0.055) / 1.055, 2.4)


def _linear_to_srgb(l):
    if l < 0.00313066844250063:
        return l * 12.92
    return pow(l, 1 / 2.4) * 1.055 - 0.055


def _srgb_to_linear_numpy(s):
    choicelist = [s / 12.92, ((s + 0.055) / 1.055) ** 2.4]
    return numpy.select([s < 0.0404482362771082, True], choicelist)


def _linear_to_srgb_numpy(l):
    choicelist = [l * 12.92, (l ** (1 / 2.4)) * 1.055 - 0.055]
    return numpy.select([l < 0.00313066844250063, True], choicelist)


def _rgb_to_hsv(r, g, b):
    max_v = v = max(r, g, b)
    min_v = min(r, g, b)
    d = max_v - min_v
    s = 0 if max_v == 0 else d / max_v
    if max_v == min_v:
        h = 0
    elif max_v == r:
        h = (g - b) / d + (6 if g < b else 0)
    elif max_v == g:
        h = (b - r) / d + 2
    else:  # max_v == b
        h = (r - g) / d + 4
    h /= 6
    return h, s, v


def _hsv_to_rgb(h, s, v):
    i = int(h * 6)
    f = h * 6 - i
    p = v * (1 - s)
    q = v * (1 - f * s)
    t = v * (1 - (1 - f) * s)
    if i == 0: return v, t, p
    if i == 1: return q, v, p
    if i == 2: return p, v, t
    if i == 3: return p, q, v
    if i == 4: return t, p, v
    return v, p, q  # if i == 5:


def _rgb_to_yuv(r, g, b):
    y = (0.299 * r) + (0.587 * g) + (0.114 * b)
    u = (1.0 / 1.772) * (b - y)
    v = (1.0 / 1.402) * (r - y)
    return y, u, v


def _yuv_to_rgb(y, u, v):
    r = 1.402 * v + y
    g = (y - (0.299 * 1.402 / 0.587) * v - (0.114 * 1.772 / 0.587) * u)
    b = 1.772 * u + y
    return r, g, b


def rgb_color_enhance(source,
                      brightness=0, exposure=0, contrast=0, warmth=0,
                      saturation=0, vibrance=0,
                      # highlights=0, shadows=0,
                      hue=0, gamma=1.0,
                      linear=False, cls=ImageFilter.Color3DLUT):
    """Generates 3D color lookup table based on given values of basic
    color settings.

    :param source: Could be the source lookup table which will be modified,
                   or just a size of new identity table, from 2 to 65.
                   Performance can dramatically decrease wth bigger sizes.
    :param brightness: One value for all channels, or tuple of three values
                       from -1.0 to 1.0. Use ``exposure`` for better result.
    :param exposure: One value for all channels, or tuple of three values
                     from -5.0 to 5.0.
    :param contrast: One value for all channels, or tuple of three values
                     from -1.0 to 5.0.
    :param warmth: One value from -1.0 to 1.0.
    :param saturation: One value for all channels, or tuple of three values
                       from -1.0 to 5.0.
    :param vibrance: One value for all channels, or tuple of three values
                     from -1.0 to 5.0.
    :param hue: One value from 0 to 1.0.
    :param gamma: One value from 0 to 10.0. Default is 1.0.
    :param linear: boolean value. Convert values from sRGB to linear color space
                   before the manipulating and return after. Default is False.
                   Most arguments more sensitive in this mode.
    """
    source_is_lut = hasattr(source, 'table')
    if source_is_lut and source.channels != 3:
        raise ValueError("Only 3-channels table could be a source")

    if brightness:
        if not isinstance(brightness, (tuple, list)):
            brightness = (brightness, brightness, brightness)
        if any(not -1.0 <= x <= 1.0 for x in brightness):
            raise ValueError("Brightness should be from -1.0 to 1.0")

    if exposure:
        if not isinstance(exposure, (tuple, list)):
            exposure = (exposure, exposure, exposure)
        if any(not -5.0 <= x <= 5.0 for x in exposure):
            raise ValueError("Exposure should be from -5.0 to 5.0")
        exposure = [(2**x) for x in exposure]

    if contrast:
        if not isinstance(contrast, (tuple, list)):
            contrast = (contrast, contrast, contrast)
        if any(not -1.0 <= x <= 5.0 for x in contrast):
            raise ValueError("Contrast should be from -1.0 to 5.0")
        contrast = [(x + 1) for x in contrast]

    if warmth:
        if not -1.0 <= warmth <= 1.0:
            raise ValueError("Warmth should be from -1.0 to 1.0")
        if warmth < 0:
            warmth = (warmth * -0.0588, warmth * -0.1569, warmth * 0.1255)
        else:
            warmth = (warmth * 0.1765, warmth * -0.1255, warmth * 0.0902)

    if saturation:
        if not isinstance(saturation, (tuple, list)):
            saturation = (saturation, saturation, saturation)
        if any(not -1.0 <= x <= 5.0 for x in saturation):
            raise ValueError("Saturation should be from -1.0 to 5.0")

    if vibrance:
        if not isinstance(vibrance, (tuple, list)):
            vibrance = (vibrance, vibrance, vibrance)
        if any(not -1.0 <= x <= 5.0 for x in vibrance):
            raise ValueError("Vibrance should be from -1.0 to 1.0")
        vibrance = [(x * 2) for x in vibrance]

    # if highlights:
    #     if not isinstance(highlights, (tuple, list)):
    #         highlights = (highlights, highlights, highlights)
    #     if any(not -1.0 <= x <= 1.0 for x in highlights):
    #         raise ValueError("Highlights should be from -1.0 to 1.0")

    # if shadows:
    #     if not isinstance(shadows, (tuple, list)):
    #         shadows = (shadows, shadows, shadows)
    #     if any(not -1.0 <= x <= 1.0 for x in shadows):
    #         raise ValueError("Shadows should be from -1.0 to 1.0")

    if not 0 <= hue <= 1.0:
        raise ValueError("Hue should be from 0.0 to 1.0")

    if gamma != 1:
        if not isinstance(gamma, (tuple, list)):
            gamma = (gamma, gamma, gamma)
        if any(not 0 <= x <= 10 for x in gamma):
            raise ValueError("Gamma should be from 0.0 to 10.0")

    if numpy and not hue:
        if source_is_lut:
            size = source.size
            points = numpy.array(source.table, copy=False, dtype=numpy.float32)
            r = points[0::3]
            g = points[1::3]
            b = points[2::3]
        else:
            size = cls._check_size(source)
            b, g, r = numpy.mgrid[
                0 : 1 : size[2]*1j,
                0 : 1 : size[1]*1j,
                0 : 1 : size[0]*1j
            ].astype(numpy.float32)

        if linear:
            r = _srgb_to_linear_numpy(r)
            g = _srgb_to_linear_numpy(g)
            b = _srgb_to_linear_numpy(b)

        if contrast:
            r = (r - 0.5) * contrast[0] + 0.5
            g = (g - 0.5) * contrast[1] + 0.5
            b = (b - 0.5) * contrast[2] + 0.5

        if saturation:
            avg_v = r * 0.2126 + g * 0.7152 + b * 0.0722
            r += (r - avg_v) * saturation[0]
            g += (g - avg_v) * saturation[1]
            b += (b - avg_v) * saturation[2]

        if vibrance:
            max_v = numpy.maximum.reduce([r, g, b])
            avg_v = r * 0.2126 + g * 0.7152 + b * 0.0722
            r += (r - max_v) * (max_v - avg_v) * vibrance[0]
            g += (g - max_v) * (max_v - avg_v) * vibrance[1]
            b += (b - max_v) * (max_v - avg_v) * vibrance[2]

        if exposure:
            r = 1.0 - (1.0 - r).clip(0) ** exposure[0]
            g = 1.0 - (1.0 - g).clip(0) ** exposure[1]
            b = 1.0 - (1.0 - b).clip(0) ** exposure[2]

        if brightness:
            r += brightness[0]
            g += brightness[1]
            b += brightness[2]

        if warmth:
            y, u, v = _rgb_to_yuv(r, g, b)
            scale = numpy.sin(y * 3.14159)
            y += scale * warmth[0]
            u += scale * warmth[1]
            v += scale * warmth[2]
            r, g, b = _yuv_to_rgb(y, u, v)

        if gamma != 1:
            r = r.clip(0) ** gamma[0]
            g = g.clip(0) ** gamma[1]
            b = b.clip(0) ** gamma[2]

        if linear:
            r = _linear_to_srgb_numpy(r)
            g = _linear_to_srgb_numpy(g)
            b = _linear_to_srgb_numpy(b)

        table = numpy.stack((r, g, b), axis=-1)
        return cls(size, table.reshape(table.size), _copy_table=False)

    def generate(r, g, b):
        if linear:
            r = _srgb_to_linear(r)
            g = _srgb_to_linear(g)
            b = _srgb_to_linear(b)

        if contrast:
            r = (r - 0.5) * contrast[0] + 0.5
            g = (g - 0.5) * contrast[1] + 0.5
            b = (b - 0.5) * contrast[2] + 0.5

        if saturation:
            avg_v = r * 0.2126 + g * 0.7152 + b * 0.0722
            r += (r - avg_v) * saturation[0]
            g += (g - avg_v) * saturation[1]
            b += (b - avg_v) * saturation[2]

        if vibrance:
            max_v = max(r, g, b)
            avg_v = r * 0.2126 + g * 0.7152 + b * 0.0722
            r += (r - max_v) * (max_v - avg_v) * vibrance[0]
            g += (g - max_v) * (max_v - avg_v) * vibrance[1]
            b += (b - max_v) * (max_v - avg_v) * vibrance[2]

        if exposure:
            r = 1.0 - max(0, 1.0 - r) ** exposure[0]
            g = 1.0 - max(0, 1.0 - g) ** exposure[1]
            b = 1.0 - max(0, 1.0 - b) ** exposure[2]

        if brightness:
            r += brightness[0]
            g += brightness[1]
            b += brightness[2]

        if warmth:
            y, u, v = _rgb_to_yuv(r, g, b)
            scale = sin(y * 3.14159)
            y += scale * warmth[0]
            u += scale * warmth[1]
            v += scale * warmth[2]
            r, g, b = _yuv_to_rgb(y, u, v)

        if hue:
            h, s, v = _rgb_to_hsv(r, g, b)
            r, g, b = _hsv_to_rgb((h + hue) % 1, s, v)

        if gamma != 1:
            r = max(0, r) ** gamma[0]
            g = max(0, g) ** gamma[1]
            b = max(0, b) ** gamma[2]

        if linear:
            r = _linear_to_srgb(r)
            g = _linear_to_srgb(g)
            b = _linear_to_srgb(b)

        return r, g, b

    if source_is_lut:
        return source.transform(generate)
    else:
        return cls.generate(source, generate)


def identity_table(size, target_mode=None, cls=ImageFilter.Color3DLUT):
    """Returns noop lookup table with linear distributed values.

    :param size: Size of the table. From 2 to 65.
    :param target_mode: A mode for the result image. Should have not less
                        than ``channels`` channels. Default is ``None``,
                        which means that mode wouldn't be changed.
    """
    if numpy:
        size = cls._check_size(size)
        b, g, r = numpy.mgrid[
            0 : 1 : size[2]*1j,
            0 : 1 : size[1]*1j,
            0 : 1 : size[0]*1j
        ].astype(numpy.float32)

        table = numpy.stack((r, g, b), axis=-1)
        return cls(size, table.reshape(table.size),
                   target_mode=target_mode, _copy_table=False)

    return cls.generate(size, lambda r, g, b: (r, g, b),
                        target_mode=target_mode)



from __future__ import division, unicode_literals, absolute_import

try:
    from PIL import Image, ImageFilter, ImageMath
except ImportError:  # pragma: no cover
    raise ImportError("Pillow is not installed. "
                      "Please install latest Pillow or Pillow-SIMD package.")

if not hasattr(ImageFilter, "Color3DLUT"):  # pragma: no cover
    raise ImportError("Pillow with the color LUT transformations is required. "
                      "Please install latest Pillow or Pillow-SIMD package.")

from .loaders import load_cube_file, load_hald_image
from .generators import identity_table, rgb_color_enhance
from .operations import (sample_lut_linear, sample_lut_cubic, resize_lut,
                         transform_lut, amplify_lut)



from __future__ import division, unicode_literals, absolute_import

import warnings

from . import Image, ImageFilter


try:
    import numpy
except ImportError:  # pragma: no cover
    numpy = None


def _inter_linear(d, v0, v1):
    return v0 + (v1 - v0) * d


def _inter_linear_table(d, c, table, i0, i1):
    return [
        _inter_linear(d, table[i0+i], table[i1+i])
        for i in range(c)
    ]


def _inter_linear_vector(d, c, v0, v1):
    return [
        _inter_linear(d, v0[i], v1[i])
        for i in range(c)
    ]


def _inter_cubic_inner(d, v0, v1, v2, v3):
    """
    :param d: Distance from v1 to the side of v2, from 0.0 to 1.0
    """
    # https://en.wikipedia.org/wiki/Bicubic_interpolation#Bicubic_convolution_algorithm
    def filter_1l(x):
        a = -0.5
        return ((a + 2.0) * x - (a + 3.0)) * x*x + 1

    def filter_2l(x):
        a = -0.5
        return (((x - 5) * x + 8) * x - 4) * a

    return (v0 * filter_2l(1+d) +
            v1 * filter_1l(0+d) +
            v2 * filter_1l(1-d) +
            v3 * filter_2l(2-d))


def _inter_cubic(d, v0, v1, v2, v3):
    if d < 0:
        if d < -1.0:
            return _inter_linear(d + 1.0, v0, v1)
        return _inter_cubic_inner(d + 1.0, v0 * 2 - v1, v0, v1, v2)
    if d >= 1.0:
        if d >= 2.0:
            return _inter_linear(d - 1.0, v2, v3)
        return _inter_cubic_inner(d - 1.0, v1, v2, v3, v3 * 2 - v2)
    return _inter_cubic_inner(d, v0, v1, v2, v3)


def _inter_cubic_table(d, c, table, i0, i1, i2, i3):
    return [
        _inter_cubic(d, table[i0+i], table[i1+i], table[i2+i], table[i3+i])
        for i in range(c)
    ]


def _inter_cubic_vector(d, c, v0, v1, v2, v3):
    return [
        _inter_cubic(d, v0[i], v1[i], v2[i], v3[i])
        for i in range(c)
    ]


def _point_shift(size, point, left, right):
    size1D, size2D, size3D = size

    index1D = point[0] * (size1D - 1)
    index2D = point[1] * (size2D - 1)
    index3D = point[2] * (size3D - 1)
    idx1D = max(left, min(size1D - 1 - right, int(index1D)))
    idx2D = max(left, min(size2D - 1 - right, int(index2D)))
    idx3D = max(left, min(size3D - 1 - right, int(index3D)))
    shift1D = index1D - idx1D
    shift2D = index2D - idx2D
    shift3D = index3D - idx3D
    idx = idx1D + idx2D * size1D + idx3D * size1D * size2D
    return idx, shift1D, shift2D, shift3D


def _points_shift_numpy(size, points, left, right):
    size1D, size2D, size3D = size

    index1D = points[:, 0] * (size1D - 1)
    index2D = points[:, 1] * (size2D - 1)
    index3D = points[:, 2] * (size3D - 1)
    idx1D = index1D.astype(numpy.int32).clip(left, size1D - 1 - right)
    idx2D = index2D.astype(numpy.int32).clip(left, size2D - 1 - right)
    idx3D = index3D.astype(numpy.int32).clip(left, size3D - 1 - right)
    shift1D = numpy.subtract(index1D, idx1D, index1D).reshape(idx1D.shape[0], 1)
    shift2D = numpy.subtract(index2D, idx2D, index2D).reshape(idx2D.shape[0], 1)
    shift3D = numpy.subtract(index3D, idx3D, index3D).reshape(idx3D.shape[0], 1)
    idx = idx1D + idx2D * size1D + idx3D * size1D * size2D
    return idx, shift1D, shift2D, shift3D


def _sample_lut_linear_numpy(lut, points):
    s1D, s2D, s3D = lut.size
    s12D = s1D * s2D

    idx, shift1D, shift2D, shift3D = _points_shift_numpy(lut.size, points, 0, 1)
    table = numpy.array(lut.table, copy=False, dtype=numpy.float32)
    table = table.reshape(s1D * s2D * s3D, lut.channels)

    return _inter_linear(shift3D,
        _inter_linear(shift2D,
            _inter_linear(shift1D, table[idx + 0], table[idx + 1]),
            _inter_linear(shift1D, table[idx + s1D + 0],
                          table[idx + s1D + 1]),
        ),
        _inter_linear(shift2D,
            _inter_linear(shift1D, table[idx + s12D + 0],
                          table[idx + s12D + 1]),
            _inter_linear(shift1D, table[idx + s12D + s1D + 0],
                          table[idx + s12D + s1D + 1]),
        ),
    )


def sample_lut_linear(lut, point):
    """Computes the new point value from given 3D lookup table
    using linear interpolation.

    :param lut: Lookup table, ``ImageFilter.Color3DLUT`` object.
    :param point: A tuple of 3 values with coordinates in the cube,
                  normalized from 0.0 to 1.0. Could be out of range.
    """
    size1D, size2D, size3D = lut.size
    c = lut.channels
    s1Dc = size1D * c
    s12Dc = size1D * size2D * c

    idx, shift1D, shift2D, shift3D = _point_shift(lut.size, point, 0, 1)
    idx *= c

    return _inter_linear_vector(shift3D, c,
        _inter_linear_vector(shift2D, c,
            _inter_linear_table(shift1D, c, lut.table,
                idx + 0, idx + c),
            _inter_linear_table(shift1D, c, lut.table,
                idx + s1Dc + 0, idx + s1Dc + c),
        ),
        _inter_linear_vector(shift2D, c,
            _inter_linear_table(shift1D, c, lut.table,
                idx + s12Dc + 0, idx + s12Dc + c),
            _inter_linear_table(shift1D, c, lut.table,
                idx + s12Dc + s1Dc + 0, idx + s12Dc + s1Dc + c),
        ),
    )


def sample_lut_cubic(lut, point):
    """Computes the new point value from given 3D lookup table
    using cubic interpolation.

    :param lut: Lookup table, ``ImageFilter.Color3DLUT`` object.
    :param point: A tuple of 3 values with coordinates in the cube,
                  normalized from 0.0 to 1.0. Could be out of range,
                  in this case, will be linear interpolated using
                  two extreme values.
    """
    size1D, size2D, size3D = lut.size
    c = lut.channels
    s1Dc = size1D * c
    s12Dc = size1D * size2D * c
    c2, s1Dc2, s12Dc2 = c * 2, s1Dc * 2, s12Dc * 2

    if size1D < 4 or size2D < 4 or size3D < 4:
        raise ValueError("Cubic interpolation requires a table of size "
                         "4 in all dimensions at least. Switching to linear.")

    idx, shift1D, shift2D, shift3D = _point_shift(lut.size, point, 1, 2)
    idx *= c

    return _inter_cubic_vector(shift3D, c,
        _inter_cubic_vector(shift2D, c,
            _inter_cubic_table(shift1D, c, lut.table,
                idx-s12Dc-s1Dc-c, idx-s12Dc-s1Dc+0,
                idx-s12Dc-s1Dc+c, idx-s12Dc-s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx-s12Dc-c, idx-s12Dc+0, idx-s12Dc+c, idx-s12Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx-s12Dc+s1Dc-c, idx-s12Dc+s1Dc+0,
                idx-s12Dc+s1Dc+c, idx-s12Dc+s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx-s12Dc+s1Dc2-c, idx-s12Dc+s1Dc2+0,
                idx-s12Dc+s1Dc2+c, idx-s12Dc+s1Dc2+c2),
        ),
        _inter_cubic_vector(shift2D, c,
            _inter_cubic_table(shift1D, c, lut.table,
                idx-s1Dc-c, idx-s1Dc+0, idx-s1Dc+c, idx-s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx-c, idx+0, idx+c, idx+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s1Dc-c, idx+s1Dc+0, idx+s1Dc+c, idx+s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s1Dc2-c, idx+s1Dc2+0, idx+s1Dc2+c, idx+s1Dc2+c2),
        ),
        _inter_cubic_vector(shift2D, c,
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc-s1Dc-c, idx+s12Dc-s1Dc+0,
                idx+s12Dc-s1Dc+c, idx+s12Dc-s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc-c, idx+s12Dc+0, idx+s12Dc+c, idx+s12Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc+s1Dc-c, idx+s12Dc+s1Dc+0,
                idx+s12Dc+s1Dc+c, idx+s12Dc+s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc+s1Dc2-c, idx+s12Dc+s1Dc2+0,
                idx+s12Dc+s1Dc2+c, idx+s12Dc+s1Dc2+c2),
        ),
        _inter_cubic_vector(shift2D, c,
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc2-s1Dc-c, idx+s12Dc2-s1Dc+0,
                idx+s12Dc2-s1Dc+c, idx+s12Dc2-s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc2-c, idx+s12Dc2+0, idx+s12Dc2+c, idx+s12Dc2+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc2+s1Dc-c, idx+s12Dc2+s1Dc+0,
                idx+s12Dc2+s1Dc+c, idx+s12Dc2+s1Dc+c2),
            _inter_cubic_table(shift1D, c, lut.table,
                idx+s12Dc2+s1Dc2-c, idx+s12Dc2+s1Dc2+0,
                idx+s12Dc2+s1Dc2+c, idx+s12Dc2+s1Dc2+c2),
        ),
    )


def resize_lut(source, target_size, interp=Image.LINEAR,
               cls=ImageFilter.Color3DLUT):
    """Resizes given lookup table to new size using interpolation.

    :param source: Source lookup table, ``ImageFilter.Color3DLUT`` object.
    :param target_size: Size of the resulting lookup table.
    :param interp: Interpolation type, ``Image.LINEAR`` or ``Image.CUBIC``.
                   Linear is default. Cubic is dramatically slower.
    """
    size1D, size2D, size3D = cls._check_size(target_size)
    if interp == Image.LINEAR:
        sample_lut = sample_lut_linear
    elif interp == Image.CUBIC:
        sample_lut = sample_lut_cubic
    else:
        raise ValueError(
            "Only Image.LINEAR and Image.CUBIC interpolations are supported")
    if interp == Image.CUBIC and any(s < 4 for s in source.size):
        sample_lut = sample_lut_linear
        interp = Image.LINEAR
        warnings.warn("Cubic interpolation requires a table of size "
                      "4 in all dimensions at least. Switching to linear.")

    if numpy and interp == Image.LINEAR:
        shape = (size1D * size2D * size3D, 3)
        b, g, r = numpy.mgrid[
            0 : 1 : size3D*1j,
            0 : 1 : size2D*1j,
            0 : 1 : size1D*1j
        ].astype(numpy.float32)
        points = numpy.stack((r, g, b), axis=-1).reshape(shape)
        points = _sample_lut_linear_numpy(source, points)

        table = points.reshape(points.size)

    else:  # Native implementation
        table = []
        for b in range(size3D):
            for g in range(size2D):
                for r in range(size1D):
                    point = (r / (size1D-1), g / (size2D-1), b / (size3D-1))
                    table.extend(sample_lut(source, point))

    return cls((size1D, size2D, size3D), table,
               channels=source.channels, target_mode=source.mode,
               _copy_table=False)


def transform_lut(source, lut, target_size=None, interp=Image.LINEAR,
                  cls=ImageFilter.Color3DLUT):
    """Transforms given lookup table using another table and returns the result.
    Sizes of the tables do not have to be the same. Moreover,
    you can set the result table size with ``target_size`` argument.

    :param source: Source lookup table, ``ImageFilter.Color3DLUT`` object.
    :param lut: Applied lookup table, ``ImageFilter.Color3DLUT`` object.
    :param target_size: Optional size of the resulting lookup table.
                        By default, size of the ``source`` will be used.
    :param interp: Interpolation type, ``Image.LINEAR`` or ``Image.CUBIC``.
                   Linear is default. Cubic is dramatically slower.
    """
    if source.channels != 3:
        raise ValueError("Can transform only 3-channel cubes")
    if interp == Image.LINEAR:
        sample_lut = sample_lut_linear
    elif interp == Image.CUBIC:
        sample_lut = sample_lut_cubic
    else:
        raise ValueError(
            "Only Image.LINEAR and Image.CUBIC interpolations are supported")

    if target_size:
        size1D, size2D, size3D = cls._check_size(target_size)
    else:
        size1D, size2D, size3D = source.size

    if interp == Image.CUBIC:
        small_lut = any(s < 4 for s in lut.size)
        small_source = any(s < 4 for s in source.size)
        if small_lut or (target_size and small_source):
            sample_lut = sample_lut_linear
            interp = Image.LINEAR
            warnings.warn("Cubic interpolation requires a table of size "
                          "4 in all dimensions at least. Switching to linear.")

    if numpy and interp == Image.LINEAR:
        shape = (size1D * size2D * size3D, source.channels)
        if target_size:
            b, g, r = numpy.mgrid[
                0 : 1 : size3D*1j,
                0 : 1 : size2D*1j,
                0 : 1 : size1D*1j
            ].astype(numpy.float32)
            points = numpy.stack((r, g, b), axis=-1).reshape(shape)
            points = _sample_lut_linear_numpy(source, points)
        else:
            points = numpy.array(source.table, copy=False, dtype=numpy.float32)
            points = points.reshape(shape)

        points = _sample_lut_linear_numpy(lut, points)
        table = points.reshape(points.size)

    else:  # Native implementation
        table = []
        index = 0
        for b in range(size3D):
            for g in range(size2D):
                for r in range(size1D):
                    if target_size:
                        point = (r / (size1D-1), g / (size2D-1), b / (size3D-1))
                        point = sample_lut(source, point)
                    else:
                        point = (source.table[index + 0],
                                 source.table[index + 1],
                                 source.table[index + 2])
                        index += 3
                    table.extend(sample_lut(lut, point))

    return cls((size1D, size2D, size3D), table,
               channels=lut.channels, target_mode=lut.mode or source.mode,
               _copy_table=False)


def amplify_lut(source, scale):
    """Amplifies given lookup table compared to identity table the same size.
    For 4-channel lookup tables the fourth channel will be unschanged.

    :param source: Source lookup table, ``ImageFilter.Color3DLUT`` object.
    :param scale: One or three floats which define the amplification strength.
                  1.0 mean no changes, 0.0 transforms to identity table.
    """
    if not isinstance(scale, (tuple, list)):
        scale = (scale, scale, scale)

    if numpy:
        size1D, size2D, size3D = source.size
        sb, sg, sr = numpy.mgrid[
            0 : 1 : size3D*1j,
            0 : 1 : size2D*1j,
            0 : 1 : size1D*1j
        ].astype(numpy.float32).reshape(3, size1D * size2D * size3D)

        points = numpy.array(source.table, dtype=numpy.float32)
        points = points.reshape(size1D * size2D * size3D, source.channels)
        points[:, 0] = sr + (points[:, 0] - sr) * scale[0]
        points[:, 1] = sg + (points[:, 1] - sg) * scale[1]
        points[:, 2] = sb + (points[:, 2] - sb) * scale[2]

        return type(source)(
            source.size, points.reshape(points.size),channels=source.channels,
            target_mode=source.mode, _copy_table=False,
        )

    def transform3(sr, sg, sb, r, g, b):
        return (sr + (r - sr) * scale[0],
                sg + (g - sg) * scale[1],
                sb + (b - sb) * scale[2])

    def transform4(sr, sg, sb, r, g, b, x):
        return (sr + (r - sr) * scale[0],
                sg + (g - sg) * scale[1],
                sb + (b - sb) * scale[2],
                x)

    if source.channels == 3:
        return source.transform(transform3, with_normals=True)
    elif source.channels == 4:
        return source.transform(transform4, with_normals=True)
    else:  # pragma: no cover
        raise ValueError("The source lut should have 3 or 4 channels")
